@@ -2,10 +2,10 @@
  *
  * pgsp_json.c: Plan handler for JSON/XML/YAML style plans
  *
- * Copyright (c) 2012-2015, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2012-2019, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  *
  * IDENTIFICATION
- *	  pg_store_plan/pgsp_json.c
+ *	  pg_store_plans/pgsp_json.c
  *
  *-------------------------------------------------------------------------
  */
@@ -97,6 +97,15 @@ word_table propfields[] =
 	{P_GroupKey,    	"-" ,"Group Key",			NULL, true,  NULL,				SETTER(group_key)},
 	{P_GroupSets,    	"=" ,"Grouping Sets",		NULL, true,  NULL,				NULL},
 	{P_GroupKeys,    	"\\" ,"Group Keys",			NULL, true,  NULL,				SETTER(group_key)},
+
+	{P_HashKeys,    	"~" ,"Hash Keys",			NULL, true,  NULL,				SETTER(hash_key)},
+	{P_HashKey,    		"|" ,"Hash Key",			NULL, true,  NULL,				SETTER(hash_key)},
+
+	{P_Parallel,    	"`" ,"Parallel Aware",		NULL, true,  NULL,				SETTER(parallel_aware)},
+	{P_PartialMode,    	">" ,"Partial Mode",		NULL, true,  conv_partialmode,SETTER(partial_mode)},
+	{P_WorkersPlanned, 	"{" ,"Workers Planned",		NULL, true,  NULL,				SETTER(workers_planned)},
+	{P_WorkersLaunched, "}" ,"Workers Launched",	NULL, true,  NULL,				SETTER(workers_launched)},
+	{P_InnerUnique,		"?" ,"Inner Unique",		NULL, true,  NULL,				SETTER(inner_unique)},
 														  
 	/* Values of these properties are ignored on normalization */
 	{P_FunctionCall,	"y" ,"Function Call",		NULL, false, NULL,				SETTER(func_call)},
@@ -142,9 +151,13 @@ word_table propfields[] =
 	{P_ConfArbitIdx,    "@" ,"Conflict Arbiter Indexes",NULL, false,  NULL,			SETTER(conflict_arbiter_indexes)},
 	{P_TuplesInserted,  "^" ,"Tuples Inserted",		NULL, false,  NULL,				SETTER(tuples_inserted)},
 	{P_ConfTuples,		"+" ,"Conflicting Tuples",	NULL, false,  NULL,				SETTER(conflicting_tuples)},
-	{P_SamplingMethod,  ""  ,"Sampling Method" ,	NULL, false,  NULL,				SETTER(sampling_method)},
-	{P_SamplingParams,  ""  ,"Sampling Parameters" , NULL, false,  NULL,			SETTER(sampling_params)},
-	{P_RepeatableSeed,  ""  ,"Repeatable Seed" ,	NULL, false,  NULL,				SETTER(repeatable_seed)},
+	{P_SamplingMethod,  ":"  ,"Sampling Method" ,	NULL, false,  NULL,				SETTER(sampling_method)},
+	{P_SamplingParams,  ";"  ,"Sampling Parameters" , NULL, false,  NULL,			SETTER(sampling_params)},
+	{P_RepeatableSeed,  "<"  ,"Repeatable Seed" ,	NULL, false,  NULL,				SETTER(repeatable_seed)},
+	{P_Workers,    		"[" ,"Workers",				NULL, false,  NULL,				NULL},
+	{P_WorkerNumber,    "]" ,"Worker Number",		NULL, false,  NULL,				SETTER(worker_number)},
+	{P_TableFuncName,    "aa" ,"Table Function Name",NULL, false,  NULL,			SETTER(table_func_name)},
+
 	{P_Invalid, NULL, NULL, NULL, false, NULL, NULL}
 };
 
@@ -184,6 +197,19 @@ word_table nodetypes[] =
 	{T_SetOp,		"3" ,"SetOp",			NULL, false, NULL, NULL},
 	{T_LockRows,	"4" ,"LockRows",		NULL, false, NULL, NULL},
 	{T_Limit,		"5" ,"Limit",			NULL, false, NULL, NULL},
+#if PG_VERSION_NUM >= 90500
+	{T_SampleScan,	"B" ,"Sample Scan",		NULL, false, NULL, NULL},
+#if PG_VERSION_NUM >= 90600
+	{T_Gather,		"6" ,"Gather",			NULL, false, NULL, NULL},
+#if PG_VERSION_NUM >= 100000
+	{T_ProjectSet,	"7" ,"ProjectSet",		NULL, false, NULL, NULL},
+	{T_TableFuncScan,"8","Table Function Scan",	NULL, false, NULL, NULL},
+	{T_NamedTuplestoreScan,"9","Named Tuplestore Scan",	NULL, false, NULL, NULL},
+	{T_GatherMerge,	"A" ,"Gather Merge",	NULL, false, NULL, NULL},
+#endif
+#endif
+#endif
+
 	{T_Invalid,		NULL, NULL, NULL, false, NULL, NULL}
 };
 
@@ -211,6 +237,7 @@ word_table strategies[] =
 	{S_Plain,	"p" ,"Plain", NULL, false, NULL, NULL},
 	{S_Sorted,	"s" ,"Sorted", NULL, false, NULL, NULL},
 	{S_Hashed,	"h" ,"Hashed", NULL, false, NULL, NULL},
+	{S_Mixed,	"m" ,"Mixed", NULL, false, NULL, NULL},
 	{S_Invalid,	NULL, NULL, NULL, false, NULL, NULL}
 };
 
@@ -259,6 +286,14 @@ word_table sortspacetype[] =
 	{T_Invalid, NULL, NULL, NULL, false, NULL, NULL}
 };
 
+word_table partialmode[] =
+{
+	{T_Invalid,  "p" ,"Partial",	NULL, false, NULL, NULL},
+	{T_Invalid,  "f" ,"Finalize",NULL, false, NULL, NULL},
+	{T_Invalid,  "s" ,"Simple",NULL, false, NULL, NULL},
+	{T_Invalid, NULL, NULL, NULL, false, NULL, NULL}
+};
+
 word_table *
 search_word_table(word_table *tbl, const char *word, int mode)
 {
@@ -363,6 +398,7 @@ conv_strategy(const char *src, pgsp_parser_mode mode)
 		    tok == TRUE_P || tok == FALSE_P || \
 			tok == CURRENT_DATE || tok == CURRENT_TIME || \
 		    tok == LOCALTIME || tok == LOCALTIMESTAMP)
+#define IS_INDENTED_ARRAY(v) ((v) == P_GroupKeys || (v) == P_HashKeys)
 
 /*
  * norm_yylex: core_yylex with replacing some tokens.
@@ -415,6 +451,11 @@ norm_yylex(char *str, core_YYSTYPE *yylval, YYLTYPE *yylloc, core_yyscan_t yysca
  * uniqueness, preserve_space puts one space for one existent whitespace for
  * more readability.
  */
+/* scanner interface is changed in PG12 */
+#if PG_VERSION_NUM < 120000
+#define ScanKeywords (*ScanKeywords)
+#define ScanKeywordTokens NumScanKeywords
+#endif
 void
 normalize_expr(char *expr, bool preserve_space)
 {
@@ -430,9 +471,16 @@ normalize_expr(char *expr, bool preserve_space)
 	wp = expr;
 	yyscanner = scanner_init(expr,
 							 &yyextra,
-							 ScanKeywords,
-							 NumScanKeywords);
+							 &ScanKeywords,
+							 ScanKeywordTokens);
 
+	/*
+	 * The warnings about nonstandard escape strings is already emitted in the
+	 * core. Just silence them here.
+	 */
+#if PG_VERSION_NUM >= 90500
+	yyextra.escape_string_warning = false;
+#endif
 	lasttok = 0;
 	lastloc = -1;
 
@@ -458,6 +506,18 @@ normalize_expr(char *expr, bool preserve_space)
 				memcpy(wp, expr + i, i2 - i);
 				wp += i2 - i;
 			}
+#if PG_VERSION_NUM >= 100000
+			/*
+			 * Since PG10 pg_stat_statements doesn't store trailing semicolon
+			 * in the column "query". Normalization is basically useless in the
+			 * version but still usefull to match utility commands so follow
+			 * the behavior change.
+			 */
+			else if (lasttok == ';')
+			{
+				/* Just do nothing */
+			}
+#endif
 			else
 			{
 				/* Upcase keywords */
@@ -591,6 +651,12 @@ conv_sortspacetype(const char *src, pgsp_parser_mode mode)
 	return converter_core(sortspacetype, src, mode);
 }
 
+const char *
+conv_partialmode(const char *src, pgsp_parser_mode mode)
+{
+	return converter_core(partialmode, src, mode);
+}
+
 /**** Parser callbacks ****/
 
 /* JSON */
@@ -641,7 +707,7 @@ json_arrstart(void *state)
 {
 	pgspParserContext *ctx = (pgspParserContext *)state;
 
-	if (ctx->current_list == P_GroupKeys)
+	if (IS_INDENTED_ARRAY(ctx->current_list))
 		ctx->wlist_level++;
 
 	appendStringInfoChar(ctx->dest, '[');
@@ -656,11 +722,12 @@ json_arrend(void *state)
 {
 	pgspParserContext *ctx = (pgspParserContext *)state;
 
-	if (ctx->current_list == P_GroupKeys)
+	if (IS_INDENTED_ARRAY(ctx->current_list))
 		ctx->wlist_level--;
 
-	if (ctx->current_list == P_GroupKeys ? ctx->wlist_level == 0 :
-		(ctx->mode == PGSP_JSON_INFLATE && ctx->last_elem_is_object))
+	if (ctx->mode == PGSP_JSON_INFLATE &&
+		(IS_INDENTED_ARRAY(ctx->current_list) ?
+		 ctx->wlist_level == 0 : ctx->last_elem_is_object))
 	{
 		appendStringInfoChar(ctx->dest, '\n');
 		appendStringInfoSpaces(ctx->dest, (ctx->level - 1) * INDENT_STEP);
@@ -725,7 +792,7 @@ json_ofstart(void *state, char *fname, bool isnull)
 	if (ctx->mode == PGSP_JSON_INFLATE)
 		appendStringInfoChar(ctx->dest, ' ');
 
-	if (p && p->tag == P_GroupKeys)
+	if (p && IS_INDENTED_ARRAY(p->tag))
 	{
 		ctx->current_list = p->tag;
 		ctx->list_fname = fname;
@@ -752,14 +819,17 @@ json_aestart(void *state, bool isnull)
 	if (ctx->remove)
 		return;
 
-	if (ctx->current_list == P_GroupKeys &&
+	if (IS_INDENTED_ARRAY(ctx->current_list) &&
 		ctx->wlist_level == 1)
 	{
 		if (!bms_is_member(ctx->level, ctx->first))
 			appendStringInfoChar(ctx->dest, ',');
-		
-		appendStringInfoChar(ctx->dest, '\n');
-		appendStringInfoSpaces(ctx->dest, (ctx->level) * INDENT_STEP);
+
+		if (ctx->mode == PGSP_JSON_INFLATE)
+		{
+			appendStringInfoChar(ctx->dest, '\n');
+			appendStringInfoSpaces(ctx->dest, (ctx->level) * INDENT_STEP);
+		}
 	}
 	else
 	{
