@@ -2,10 +2,10 @@
  *
  * pgsp_json.h: Definitions of plan handler for JSON/XML/YAML style plans
  *
- * Copyright (c) 2012-2015, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2012-2019, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  *
  * IDENTIFICATION
- *	  pg_store_plan/pgsp_json.h
+ *	  pg_store_plans/pgsp_json.h
  *
  *-------------------------------------------------------------------------
  */
