@@ -1,11 +1,11 @@
 /*-------------------------------------------------------------------------
  *
- * pgsp_json_text.h: Text plan generator for pg_store_plan.
+ * pgsp_json_text.h: Text plan generator for pg_store_plans.
  *
- * Copyright (c) 2012-2015, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2012-2019, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  *
  * IDENTIFICATION
- *	  pg_store_plan/pgsp_json_text.c
+ *	  pg_store_plans/pgsp_json_text.c
  *
  *-------------------------------------------------------------------------
  */
@@ -14,6 +14,7 @@
 #include "miscadmin.h"
 #include "nodes/nodes.h"
 #include "nodes/bitmapset.h"
+#include "nodes/pg_list.h"
 #include "utils/json.h"
 #include "utils/jsonapi.h"
 #include "utils/builtins.h"
@@ -39,6 +40,29 @@ static void json_text_ofend(void *state, char *fname, bool isnull);
 static void json_text_scalar(void *state, char *token, JsonTokenType tokentype);
 
 /* Parser callbacks for plan textization */
+
+/*
+ * This setter is used for field names that store_plans doesn't know of.
+ * Unlike the other setters, this holds a list of strings emitted as is in text
+ * explains.
+ */
+SETTERDECL(_undef)
+{
+	StringInfo s;
+
+	if(vals->_undef_newelem)
+	{
+		s = makeStringInfo();
+		vals->_undef = lappend(vals->_undef, s);
+	}
+	else
+	{
+		s = llast (vals->_undef);
+	}
+
+	appendStringInfoString(s, val);
+}
+
 SETTERDECL(node_type)
 {
 	word_table *p;
@@ -60,6 +84,9 @@ SETTERDECL(strategy)
 	
 	p = search_word_table(strategies, val, PGSP_JSON_TEXTIZE);
 
+	if (!p)
+		return;
+
 	switch (vals->nodetag)
 	{
 		case T_Agg:
@@ -69,6 +96,8 @@ SETTERDECL(strategy)
 					vals->node_type = "HashAggregate"; break;
 				case S_Sorted:
 					vals->node_type = "GroupAggregate"; break;
+				case S_Mixed:
+					vals->node_type = "MixedAggregate"; break;
 				default:
 					break;
 			}
@@ -94,6 +123,9 @@ CONVERSION_SETTER(setopcommand, conv_setsetopcommand);
 CONVERSION_SETTER(sort_method, conv_sortmethod);
 LIST_SETTER(sort_key);
 LIST_SETTER(group_key);
+LIST_SETTER(hash_key);
+BOOL_SETTER(parallel_aware);
+CONVERSION_SETTER(partial_mode, conv_partialmode);
 SQLQUOTE_SETTER(index_name);
 DEFAULT_SETTER(startup_cost);
 DEFAULT_SETTER(total_cost);
@@ -150,6 +182,11 @@ DEFAULT_SETTER(conflicting_tuples);
 DEFAULT_SETTER(sampling_method);
 LIST_SETTER(sampling_params);
 DEFAULT_SETTER(repeatable_seed);
+DEFAULT_SETTER(worker_number);
+DEFAULT_SETTER(workers_planned);
+DEFAULT_SETTER(workers_launched);
+BOOL_SETTER(inner_unique);
+DEFAULT_SETTER(table_func_name);
 
 #define ISZERO(s) (!s || strcmp(s, "0") == 0 || strcmp(s, "0.000") == 0 )
 #define HASSTRING(s) (s && strlen(s) > 0)
@@ -250,8 +287,9 @@ print_groupingsets_if_exists(StringInfo s, List *gss, int level, int exind)
 		foreach (lcg, gs->group_keys)
 		{
 			const char *gk = (const char *)lfirst (lcg);
-			print_prop_if_exists(s, "Group Key: ", gk, level, exind);
+			print_prop_if_exists(s, gs->key_type, gk, level, exind);
 		}
+
 	}
 }
 
@@ -273,7 +311,13 @@ print_current_node(pgspParserContext *ctx)
 	bool comma = false;
 	int exind = 0;
 
-	if (v->node_type == T_Invalid)
+	/*
+	 * The element objects in "Workers" list doesn't have node type, which
+	 * would be named T_Worker if there were in node.h. So it needs a special
+	 * treat.
+	 */
+	
+	if (v->node_type == T_Invalid && !HASSTRING(v->worker_number))
 		return;
 
 	if (s->len > 0)
@@ -292,6 +336,9 @@ print_current_node(pgspParserContext *ctx)
 	if (level > 1 && ctx->current_list == P_Invalid)
 		appendStringInfoString(s, "->  ");
 
+	if (v->parallel_aware)
+		appendStringInfoString(s, "Parallel ");
+
 	switch (v->nodetag)
 	{
 		case T_ModifyTable:
@@ -340,7 +387,21 @@ print_current_node(pgspParserContext *ctx)
 			break;
 
 		default:
-			appendStringInfoString(s, v->node_type);
+			/* Existence of worker_number suggests this is a Worker node */
+			if (HASSTRING(v->worker_number))
+			{
+				appendStringInfoString(s, "Worker");
+				print_prop_if_exists(s, " ", v->worker_number, 0, 0);
+
+				/* 
+				 * "Worker"s are individual JSON objects in a JSON list but
+				 * should be printed as just a property in text
+				 * representaion. Correct indent using exind here.
+				 */
+				exind = -4;
+			}
+			else
+				appendStringInfoString(s, v->node_type);
 			break;
 	}
 
@@ -405,6 +466,8 @@ print_current_node(pgspParserContext *ctx)
 	print_prop_if_exists(s, "Join Filter: " , v->join_filter, level, exind);
 	print_prop_if_exists(s, "Index Cond: " , v->index_cond, level, exind);
 	print_prop_if_exists(s, "Recheck Cond: ", v->recheck_cond, level, exind);
+	print_prop_if_exists(s, "Workers Planned: ", v->workers_planned, level, exind);
+	print_prop_if_exists(s, "Workers Launched: ", v->workers_launched, level, exind);
 
 	if (HASSTRING(v->sampling_method))
 	{
@@ -437,6 +500,21 @@ print_current_node(pgspParserContext *ctx)
 	}
 
 	print_prop_if_exists(s, "Function Call: ", v->func_call, level, exind);
+
+	/*
+	 * Emit unknown properties here. The properties are printed in the same
+	 * shape with JSON properties as assumed by explain.c.
+	 */
+	foreach (lc, v->_undef)
+	{
+		StringInfo str = (StringInfo) lfirst(lc);
+
+		appendStringInfoString(s, "\n");
+		appendStringInfoSpaces(s, TEXT_INDENT_DETAILS(level, exind));
+		appendStringInfoString(s, str->data);
+	}
+	v->_undef = NULL;
+
 	print_prop_if_exists(s, "Filter: ", v->filter, level, exind);
 	print_prop_if_nz(s, "Rows Removed by Filter: ",
 						 v->filter_removed, level, exind);
@@ -670,8 +748,11 @@ json_text_objstart(void *state)
 			v->sort_key = makeStringInfo();
 		if (!v->group_key)
 			v->group_key = makeStringInfo();
+		if (!v->hash_key)
+			v->hash_key = makeStringInfo();
 		resetStringInfo(v->sort_key);
 		resetStringInfo(v->group_key);
+		resetStringInfo(v->hash_key);
 	}
 }
 
@@ -760,11 +841,26 @@ json_text_arrend(void *state)
 			 * go into individual "Group Key" lines. Empty innermost list is
 			 * represented as "()" there. See explain.c of PostgreSQL.
 			 */
-			ctx->tmp_gset->group_keys =
-				lappend(ctx->tmp_gset->group_keys,
-						(v->group_key->data[0] ?
-						 pstrdup(v->group_key->data) : "()"));
+			ctx->tmp_gset->key_type = "Group Key: ";
+			if (v->group_key->data[0])
+			{
+				ctx->tmp_gset->group_keys =
+					lappend(ctx->tmp_gset->group_keys,
+							pstrdup(v->group_key->data));
+			}
+			else if (v->hash_key->data[0])
+			{
+				ctx->tmp_gset->group_keys =
+					lappend(ctx->tmp_gset->group_keys,
+							pstrdup(v->hash_key->data));
+				ctx->tmp_gset->key_type = "Hash Key: ";
+			}
+			else
+				ctx->tmp_gset->group_keys =
+					lappend(ctx->tmp_gset->group_keys, "()");
+
 			resetStringInfo(ctx->nodevals->group_key);
+			resetStringInfo(ctx->nodevals->hash_key);
 		}
 		ctx->wlist_level--;
 	}
@@ -781,17 +877,29 @@ json_text_ofstart(void *state, char *fname, bool isnull)
 
 	if (!p)
 	{
-		ereport(DEBUG1,
+		ereport(DEBUG2,
 				(errmsg("Short JSON parser encoutered unknown field name: \"%s\", skipped.", fname),
 				 errdetail_log("INPUT: \"%s\"", ctx->org_string)));
+
+		/*
+		 * Unknown properties may be put by foreign data wrappers and assumed
+		 * to be printed in the same format to JSON properties. We store in
+		 * nodevals a string emittable as-is in text explains.
+		 */
+		ctx->setter = SETTER(_undef);
+		ctx->nodevals->_undef_newelem = true;
+		ctx->setter(ctx->nodevals, fname);
+		ctx->nodevals->_undef_newelem = false;
+		ctx->setter(ctx->nodevals, ": ");
 	}
 	else
 	{
 		/*
-		 * Print node immediately if the next level of Plan/Plans comes. The
-		 * plan construct is tail-recursive so this doesn't harm.
+		 * Print the current node immediately if the next level of
+		 * Plan/Plans/Worers comes. This assumes that the plan output is
+		 * strcutured tail-recursively.
 		 */
-		if (p->tag == P_Plan || p->tag == P_Plans)
+		if (p->tag == P_Plan || p->tag == P_Plans || p->tag == P_Workers)
 		{
 			print_current_node(ctx);
 			clear_nodeval(ctx->nodevals);
@@ -808,7 +916,8 @@ json_text_ofstart(void *state, char *fname, bool isnull)
 			v->tmp_schema_name = v->schema_name;
 			v->tmp_alias = v->alias;
 		}
-		else if (p->tag == P_GroupSets)
+
+		if (p->tag == P_GroupSets || p->tag == P_Workers)
 		{
 			ctx->current_list = p->tag;
 			ctx->list_fname = fname;
@@ -819,7 +928,7 @@ json_text_ofstart(void *state, char *fname, bool isnull)
 		 * This paser prints partial result at the end of every P_Plan object,
 		 * which includes elements in P_Plans list.
 		 */
-		if (p->tag == P_Plan || p->tag == P_Plans)
+		if (p->tag == P_Plan || p->tag == P_Plans || p->tag == P_Workers)
 			ctx->plan_levels = bms_add_member(ctx->plan_levels, ctx->level);
 		else
 			ctx->plan_levels = bms_del_member(ctx->plan_levels, ctx->level);
